from yta_general_utils.programming.parameter_validator import PythonValidator
from typing import Union
from dataclasses import dataclass

import inspect


@dataclass
class MethodParameters:
    """
    Dataclass to hold the parameters of a method,
    classified in mandatory or optional according
    to its declaration.

    Those parameters are obtained from the method
    signature, where any parameter with a default
    None value is considered optional.
    """

    mandatory_parameters: list[str]
    """
    The names of those parameters that have been
    considered mandatory according to the method
    signature.
    """
    optional_parameters: list[str]
    """
    The names of those parameters that have been
    considered optional according to the method
    signature.
    """

    def __init__(
        self,
        mandatory_parameters: list[str],
        optional_parameters: list[str]
    ):
        self.mandatory_parameters = mandatory_parameters
        self.optional_parameters = optional_parameters

@dataclass
class Parameter:
    """
    The representation of a parameter found in a
    method signature, indicating if it was what 
    we consider an optional parameter or a 
    mandatory one based on its default value.
    """

    name: str
    """
    The name of the parameter.
    """
    type: any
    """
    The type of the parameter.
    """
    default_value: any
    """
    The default value of the parameter.
    """

    def __init__(
        self,
        name: str,
        type: any,
        default_value: any
    ):
        self.name = name
        # TODO: The type, that comes from '.annotation' has
        # to be reconsidered as it is quite special
        self.type = type
        self.default_value = default_value

    @property
    def is_default_value_empty(
        self
    ) -> bool:
        """
        Check if the default value is empty (no value
        associated) or  not.
        """
        return self.default_value is inspect._empty
    
    @property
    def is_optional(
        self
    ) -> bool:
        """
        Check if the parameter is optional (its default
        value is None).
        """
        return self.default_value is None
    
    @property
    def is_mandatory(
        self
    ) -> bool:
        """
        Check if the parameter is mandatory (its default
        value is not None).
        """
        return not self.is_optional
    
@dataclass
class Parameters:
    """
    List of parameters that exist in a function
    or method signature, to be able to handle
    them easy and detect which ones are mandatory
    or optional and more stuff.
    """

    def __init__(
        self,
        parameters: Union[dict[str, Parameter], list[Parameter]]
    ):
        if (
            PythonValidator.is_list(parameters) and
            all(PythonValidator.is_instance(item, Parameter) for item in parameters)
        ):
            parameters = {
                parameter.name: parameter
                for parameter in parameters
            }

        # TODO: Validate, in a else, if not dict[str, Parameter]
        self.parameters = parameters

    @property
    def mandatory(
        self
    ) -> dict[str, Parameter]:
        """
        Get the parameters that are mandatory.
        """
        return {
            key: value
            for key, value in self.parameters.items()
            if value.is_mandatory
        }
    
    @property
    def optional(
        self
    ) -> dict[str, Parameter]:
        """
        Get the parameters that are optional.
        """
        return {
            key: value
            for key, value in self.parameters.items()
            if value.is_optional
        }
